/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: video_out_xshm.c,v 1.20 2000/11/16 23:21:27 f1rmb Exp $
 * 
 * video_out_xshm.c, X11 shared memory extension interface for xine
 *
 * based on mpeg2dec code from
 * Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * xine-specific code by Guenter Bartsch <bartscgr@studbox.uni-stuttgart.de>
 * 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "video_out.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <errno.h>

#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>

#include "../libmpeg2/mpeg2.h"
#include "yuv2rgb.h"
#include "../gui/gui_dnd.h"

/*
#define VERBOSE
*/

extern Display *gDisplay;
extern Window   gVideoWin;
extern Pixmap   gXineLogo;
extern int      gXineLogoWidth, gXineLogoHeight;

typedef struct xshm_image_info_s {
  vo_image_buffer_t  mImageBuffer;
  XImage            *mImage;
  XShmSegmentInfo    mShminfo;
} xshm_image_buffer_t;

typedef struct _xshm_globals {
  int             screen;
  Window          window;
  XVisualInfo     vinfo;
  int             depth,bpp;
  int             bFullscreen;
  int             bIsFullscreen;
  GC              gc;
  uint32_t        image_width;
  uint32_t        image_height;
  uint32_t        image_xoff;
  uint32_t        image_yoff;
  uint32_t        dest_width;
  uint32_t        dest_height;
  uint32_t        ratio;
  int             bYuvInitialized;

  /*
   * misc (read: fun ;))
   */
  int                bLogoMode;
} xshm_globals;

xshm_globals gXshm ;

#define XLOCK { XLockDisplay (gDisplay) ; }
#define XUNLOCK { XUnlockDisplay (gDisplay) ; }

static int get_capabilities_xshm () {
  return VO_CAP_COPIES_IMAGE;
}

static void setup_window_xshm () {
  char                 *hello = "xine video output";
  Colormap              theCmap;
  XSizeHints            hint;
  XEvent                xev;
  XGCValues             xgcv;
  XColor                background, ignored;
  XSetWindowAttributes  attr;
  int                   ww, wh;
  float                 aspect;
  XClassHint            *xshm_xclasshint;

  XLOCK

    
  if((xshm_xclasshint = XAllocClassHint()) != NULL) {
    xshm_xclasshint->res_name = "Xine Xshm Video";
    xshm_xclasshint->res_class = "Xine";
  }

  if (gXshm.bFullscreen) {

    ww = DisplayWidth (gDisplay, gXshm.screen);
    wh = DisplayHeight (gDisplay, gXshm.screen);
    
    /*
     * zoom to fullscreen
     */

    if (gXshm.dest_width < ww) {
      aspect = (float) gXshm.dest_width / (float) gXshm.dest_height ;

      gXshm.dest_width = ww;
      gXshm.dest_height = ww / aspect;
    }

    gXshm.image_xoff = ( ww - gXshm.dest_width) / 2;
    gXshm.image_yoff = ( wh - gXshm.dest_height) / 2;


    if (gXshm.window) {

      if (gXshm.bIsFullscreen) {
	XUNLOCK
	return;
      }

      XDestroyWindow(gDisplay, gXshm.window);
      gXshm.window = 0;

    }

    gXshm.bIsFullscreen = 1;

    /*
     * open fullscreen window
     */

    if (XAllocNamedColor (gDisplay, DefaultColormap (gDisplay, gXshm.screen), 
			  "black",
			  &background, &ignored) == 0) {
      fprintf (stderr, "Cannot allocate color black\n");
      exit(1);
    }

    attr.background_pixel  = background.pixel;
    attr.override_redirect = True;

    gXshm.window =
      XCreateWindow (gDisplay, RootWindow (gDisplay, gXshm.screen), 0, 0, 
		     ww, wh, 0, gXshm.depth,
		     CopyFromParent, DefaultVisual (gDisplay, gXshm.screen),
		     CWBackPixel|CWOverrideRedirect, &attr);

    if(xshm_xclasshint != NULL)
      XSetClassHint(gDisplay, gXshm.window, xshm_xclasshint);

    gVideoWin = gXshm.window;

  } else {

    if (gXshm.window) {

      if (gXshm.bIsFullscreen) {
	XDestroyWindow(gDisplay, gXshm.window);
	gXshm.window = 0;
      } else {
	
	XResizeWindow (gDisplay, gXshm.window, gXshm.dest_width, 
		       gXshm.dest_height);

	XUNLOCK
	return;
	
      }
    }

    gXshm.bIsFullscreen = 0;


    hint.x = 0;
    hint.y = 0;
    hint.width  = gXshm.dest_width;
    hint.height = gXshm.dest_height;
    hint.flags  = PPosition | PSize;
    
    theCmap   = XCreateColormap(gDisplay, RootWindow(gDisplay,gXshm.screen), 
			      gXshm.vinfo.visual, AllocNone);
  
    attr.background_pixel = 0;
    attr.border_pixel     = 1;
    attr.colormap         = theCmap;

    gXshm.window = XCreateWindow(gDisplay, RootWindow(gDisplay, gXshm.screen),
			       hint.x, hint.y, hint.width, hint.height, 4, 
			       gXshm.depth,CopyFromParent,gXshm.vinfo.visual,
			       CWBackPixel | CWBorderPixel |CWColormap,&attr);

    if(xshm_xclasshint != NULL)
      XSetClassHint(gDisplay, gXshm.window, xshm_xclasshint);

    gVideoWin = gXshm.window;
    
    /* Tell other applications about this window */

    XSetStandardProperties(gDisplay, gXshm.window, hello, hello, 
			   None, NULL, 0, &hint);

    gXshm.image_xoff = 0;
    gXshm.image_yoff = 0;
  }
  
  XSelectInput(gDisplay, gXshm.window, StructureNotifyMask | ExposureMask | KeyPressMask | ButtonPressMask | StructureNotifyMask);

  /* Map window. */
  
  XMapWindow(gDisplay, gXshm.window);
  
  /* Wait for map. */
  do  {
    XMaskEvent(gDisplay, 
	       StructureNotifyMask, 
	       &xev) ;
  } while (xev.type != MapNotify || xev.xmap.event != gXshm.window);

  XFlush(gDisplay);
  XSync(gDisplay, False);
  
  gXshm.gc = XCreateGC(gDisplay, gXshm.window, 0L, &xgcv);

  if (gXshm.bFullscreen)
    XSetInputFocus (gDisplay, gXshm.window, RevertToNone, CurrentTime);

  /* drag and drop */

  gui_make_window_dnd_aware (gXshm.window);

  XUNLOCK
}

/* allocates ximages if necessary */
static int set_size_xshm (uint32_t width, uint32_t height, uint32_t ratio) {

  if ( (gXshm.image_width == width) 
       && (gXshm.image_height == height)
       && (gXshm.ratio == ratio) 
       && (gXshm.bFullscreen == gXshm.bIsFullscreen))
    return 0;

  gXshm.image_width  = width;
  gXshm.image_height = height;
  gXshm.ratio        = ratio;

  /*
   * calculate dest size from ratio
   */

  /*
   * Mpeg-2:
   */

  switch (gXshm.ratio) {
  case 0: /* forbidden */
    fprintf (stderr, "invalid ratio\n");
    exit (1);
    break;
  case 1: /* square */
    gXshm.dest_width = width;
    gXshm.dest_height = height;
    break;
  case 2: /* 4:3 */
    gXshm.dest_width = width ;
    gXshm.dest_height = height;
    break;
  case 3: /* 16:9 */
    gXshm.dest_width = width / 0.7031;
    gXshm.dest_height = height;
    break;
  default:
    gXshm.dest_width = width;
    gXshm.dest_height = height;
    fprintf (stderr, "invalid ratio\n");
    /*exit (1); */
    break;
  }
    
  printf ("picture size : %d x %d (Ratio: %d)\n",width,height,ratio);

  setup_window_xshm () ;

  return 1;
}

static void dispose_image_buffer_xshm (vo_image_buffer_t *vo_img) {

  xshm_image_buffer_t *img = (xshm_image_buffer_t *) vo_img;

  XLOCK

  XShmDetach(gDisplay, &img->mShminfo);
  
  XDestroyImage(img->mImage); 
  shmdt(img->mShminfo.shmaddr);
  shmctl (img->mShminfo.shmid, IPC_RMID,NULL);

  XUNLOCK
}


static vo_image_buffer_t *alloc_image_buffer_xshm () {

  uint32_t               image_size;
  xshm_image_buffer_t   *img ;

  XLOCK

  img = (xshm_image_buffer_t *) malloc (sizeof (xshm_image_buffer_t));

  if (img==NULL) {
    printf ("out of memory while allocating image buffer\n");
    exit (1);
  }

  img->mImage = XShmCreateImage(gDisplay, 
				gXshm.vinfo.visual,
				gXshm.vinfo.depth,
				ZPixmap, NULL,
				&img->mShminfo,
				gXshm.image_width, 
				gXshm.image_height);
  
  if (img->mImage == NULL )  {
    fprintf(stderr, "Shared memory error when allocating image => exit (Ximage error)\n");
    exit (1);
  }
  
  gXshm.bpp = img->mImage->bits_per_pixel;
  img->mShminfo.shmid=shmget(IPC_PRIVATE, 
			     img->mImage->bytes_per_line * img->mImage->height, 
			     IPC_CREAT | 0777);
  
  if (img->mShminfo.shmid < 0 ) {
    printf("%s: allocating image \n",strerror(errno));
    exit (1);
  }
  
  img->mShminfo.shmaddr  = (char *) shmat(img->mShminfo.shmid, 0, 0);
  
  if (img->mShminfo.shmaddr == ((char *) -1)) {
    fprintf(stderr, "Shared memory error (address error) when allocating image \n");
    exit (1);
  }
  
  img->mShminfo.readOnly = False;
  img->mImage->data = img->mShminfo.shmaddr;
  
  XShmAttach(gDisplay, &img->mShminfo);
  
  XSync(gDisplay, False);
  shmctl(img->mShminfo.shmid, IPC_RMID, 0);
  
  image_size = gXshm.image_width * gXshm.image_height;
  img->mImageBuffer.mem[0] = malloc(image_size);
  img->mImageBuffer.mem[1] = malloc(image_size/4);
  img->mImageBuffer.mem[2] = malloc(image_size/4);
  
  if (!gXshm.bYuvInitialized) {
    yuv2rgb_init((gXshm.depth == 24) ? gXshm.bpp : gXshm.depth,MODE_RGB);
    gXshm.bYuvInitialized = 1;
  }

  XUNLOCK

  return (vo_image_buffer_t *) img;
}

static void process_frame_portion_xshm (vo_image_buffer_t *img, 
					uint8_t *py, uint8_t *pu, uint8_t *pv,
					int slice_num,
					int offset){

  uint8_t *dst; 
  uint32_t subslice_offset, slice_offset; 
  xshm_image_buffer_t *xshm_img = (xshm_image_buffer_t *) img;

  slice_offset      = slice_num * 16 * gXshm.image_width;
  subslice_offset   = slice_offset + offset ; 


  dst = xshm_img->mImage->data + subslice_offset * (gXshm.bpp/8) ; 

  yuv2rgb(dst, 
	  py, pu, pv,
	  16, 16,
	  gXshm.image_width*(gXshm.bpp/8), gXshm.image_width, gXshm.image_width/2);
}

static int is_fullscreen_xshm () {
  return gXshm.bFullscreen;
}

static void set_fullscreen_xshm (int bFullscreen) {

  fprintf(stderr, "(!)Fullscreen not implemented for xshm.\n");

  return ; /* FIXME - not implemented for xshm */

  gXshm.bFullscreen = bFullscreen;

  set_size_xshm (gXshm.image_width, gXshm.image_height, gXshm.ratio);
}

static void display_frame_xshm (vo_image_buffer_t *vo_img) {

  xshm_image_buffer_t *img = (xshm_image_buffer_t *) vo_img;

  XLOCK

  XShmPutImage(gDisplay, gXshm.window, gXshm.gc, img->mImage, 
	       0, 0, 0, 0, 
	       img->mImage->width, img->mImage->height, True); 

  XFlush(gDisplay);

  XUNLOCK
}

static void draw_logo_xshm () {

  XLOCK

  XClearWindow (gDisplay, gXshm.window); 
  XCopyArea (gDisplay, gXineLogo, gXshm.window, gXshm.gc, 0, 0,
	     gXineLogoWidth, gXineLogoHeight, 
	     (gXshm.dest_width - gXineLogoWidth)/2, 
	     (gXshm.dest_height - gXineLogoHeight)/2);

  XUNLOCK
}

static void handle_event_xshm (XEvent *event) {

  switch (event->type) {
  case Expose:
    if (event->xexpose.window == gXshm.window) {
      if (gXshm.bLogoMode) {
	draw_logo_xshm ();
      } else {
	/* FIXME : draw current frame */
      }
    }
    break;

  }
}

static void set_logo_mode_xshm (int bLogoMode) {
  gXshm.bLogoMode = bLogoMode;

  if (bLogoMode)
    draw_logo_xshm ();
}

static void reset_xshm () {
  /* nothing */
}

static vo_driver_t vo_xshm = {
  get_capabilities_xshm,
  set_size_xshm,
  alloc_image_buffer_xshm,
  dispose_image_buffer_xshm,
  process_frame_portion_xshm,
  display_frame_xshm,
  set_fullscreen_xshm,
  is_fullscreen_xshm,
  handle_event_xshm,
  set_logo_mode_xshm,
  reset_xshm
};

/* 
 * connect to server, create and map window,
 * allocate colors and (shared) memory
 */

vo_driver_t *init_video_out_xshm () {

  int                  completionType = -1;
  XWindowAttributes    attribs;

  XLOCK

  gXshm.image_width=720;
  gXshm.image_height=576;

  /*
   * ok, X11, prepare for some video!
   */

  gXshm.screen = DefaultScreen(gDisplay);

  XGetWindowAttributes(gDisplay, DefaultRootWindow(gDisplay), &attribs);

  /*
   *
   * depth in X11 terminology land is the number of bits used to
   * actually represent the colour.
   *
   * bpp in X11 land means how many bits in the frame buffer per
   * pixel. 
   *
   * ex. 15 bit color is 15 bit depth and 16 bpp. Also 24 bit
   *     color is 24 bit depth, but can be 24 bpp or 32 bpp.
   */
  
  gXshm.depth = attribs.depth;
  
  if (gXshm.depth != 15 && gXshm.depth != 16 && gXshm.depth != 24 && gXshm.depth != 32)  {
    /* The root window may be 8bit but there might still be
     * visuals with other bit depths. For example this is the 
     * case on Sun/Solaris machines.
     */
    gXshm.depth = 24;
  }

  XMatchVisualInfo(gDisplay, gXshm.screen, gXshm.depth, TrueColor, &gXshm.vinfo);

  /*
   * check for X shared memory support
   */

  if (!XShmQueryExtension(gDisplay)) {
    fprintf(stderr, "Shared memory not supported\n\n");
    return NULL;
  }

  completionType = XShmGetEventBase(gDisplay) + ShmCompletion;

  /*
   * init global variables
   */

  gXshm.bFullscreen     = 0;
  gXshm.bIsFullscreen   = 0;
  gXshm.ratio           = 0;
  gXshm.bLogoMode       = 1;
  gXshm.bYuvInitialized = 0;

  set_size_xshm (720,576,1);

  /* 
   * If gVo.depth is 24 then it may either be a 3 or 4 byte per pixel
   * format. We can't use bpp because then we would lose the 
   * distinction between 15/16bit gVo.depth (2 byte formate assumed).
   *
   * FIXME - change yuv2rgb_init to take both gVo.depth and bpp
   * parameters
   */

  if (gXshm.depth>16) 
    printf ("\n\nWARNING: current display depth is %d. For better performance\na depth of 16 bpp is recommended!\n\n",
	    gXshm.depth);

  /*  yuv2rgb_init((gXshm.depth == 24) ? gXshm.bpp : gXshm.depth,MODE_RGB); */

  XUNLOCK

  return &vo_xshm;
}

